//    Copyright 2018 Quirino Brizi [quirino.brizi@gmail.com]
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.




const Helper =require('./helper');

class Response {

  /**
   * Create a new response object
   * 
   * @param {any} payload the payload to return to the client
   * 
   * @private
   */
  constructor(payload) {
    this.body = payload;
    this.contentType = 'application/json';
    this.headers = {};
  }

  /**
   * Allow to set the response contentType, assumed JSON by default.
   * @param {String} contentType the content type for the response
   */
  setContentType(contentType) {
    this.contentType = contentType;
    return this;
  }

  /**
   * Allows to set the status code
   * @param {Number} statusCode the status code sent to the client
   * 
   * **This method is not supposed to be used directly by the user
   * and is reserved to the framework only.**
   */
  setStatusCode(statusCode) {
    this.statusCode = statusCode;
    return this;
  }

  /**
   * Allow to add an header to the response
   * @param {String} name the header name
   * @param {String} value the header value
   */
  addHeader(name, value) {
    this.headers[name] = value;
    return this;
  }

  /**
   * Propagate the response to the client.
   * 
   * @param {any} res the express response object
   * 
   * **This method is not supposed to be used directly by the user 
   * and is reserved to the framework only.**
   */
  propagate(res) {
    res.status(Number(this.statusCode))
      .type(this.contentType);
    Object.entries(this.headers).forEach(entry => res.set(entry[0], entry[1]));
    res.send(Helper.format(this.body, this.contentType));
  }

  /**
   * Create a new response object
   * 
   * @param {any} payload the payload to return to the client
   * @returns {Response}
   *
   * @static
   * @public
   */
  static build(payload) {
    return new Response(payload);
  }

  /**
   * @deprecated
   * This method is now deprecated in favour of the Response#build method and will be removed in future releases.
   * 
   * Create a new response object
   * 
   * @param {Number} status the status code to return to the client
   * @param {any} payload the payload to return to the client
   * @returns {Response}
   * 
   * @static
   * @public
   */
  static builder(_status, payload) {
    return new Response(payload);
  }

}

/**
 * The response that is returned from the API handler.
 * @typedef Response
 * @public 
 */
module.exports = Response;