//    Copyright 2018 Quirino Brizi [quirino.brizi@gmail.com]
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

'use strict';

const apicache = require('apicache-plus');

const Helper = require('./helper');
const Logger = require('dino-core').Logger;
const RouteHandler = require('./route.handler');
const ErrorHandler = require('./error.handler');
const Component = require('dino-core').Component;
const ApiValidator = require('./api.validator');
const ApplicationStateMonitor = require('./monitoring/application.state.monitor');
const CachePerformanceMonitor = require('./monitoring/cache.performance.monitor');
const CO2EmissionMonitor = require('./monitoring/co2.emission.monitor');
const BeforeHandlerMiddleware = require('./middlewares/before.handler.middleware');
const AfterHandlerMiddleware = require('./middlewares/after.handler.middleware');

const DESCRIPTOR_PROVIDER_MAP = Object.freeze({ filesystem: './openapi/filesystem.descriptor.provider' });

class RoutingConfigurer extends Component {
  constructor({ applicationContext, environment }) {
    super();

    this.server = undefined;
    this.applicationStateMonitor = undefined;

    this.environment = environment;
    this.applicationContext = applicationContext;
  }

  async postConstruct() {
    const express = this.applicationContext.require('express');
    const app = express();
    const swaggerUi = this.applicationContext.require('swagger-ui-express');
    const swaggerParser = this.applicationContext.require('swagger-parser');

    app.use(express.json());
    this.extend(app);
    const router = express.Router();
    const apis = await swaggerParser.dereference(this.resolveDescriptorProvider().provide());
    const apiValidator = new ApiValidator(this.environment);
    apiValidator.init(apis);
    const errorHandler = this.errorHandler();

    const docsPath = this.getVariable('dino:server:docs', '/api-docs');
    router.use(docsPath, swaggerUi.serve);
    router.use(docsPath, swaggerUi.setup(apis, { explorer: false }));
    this.setupMonitoringEndpointsIfRequired(router);
    const monitoringMiddlewares = this.defineTopLevelMonitorsIfRequired();

    Object.keys(apis.paths).forEach((path) => {
      const currentPath = apis.paths[path];
      Object.keys(currentPath).forEach((method) => {
        const apiPath = Helper.normalizePath(path);
        const api = currentPath[method];
        const responseValidator = apiValidator.getResponseValidator(method, path);
        const requestHandler = this.requestHandler(api, responseValidator);

        // defines middlewares proposing built-ins and concatenating eser ones
        // as per express style the user will be able to interrupt the chain just
        // sending a response, this at current stage is not a concern but a better
        // middleware structure that avoid this kind if hacks should be put in place.
        // this is tracked on https://gitlab.com/codesketch/dino-express/issues/7
        const userDefinedMiddlewares = this.middlewares(api, apis.components) || [];
        let middlewares = monitoringMiddlewares.before.concat([apiValidator.getRequestValidator(method, path)]);
        // set BeforeHandlerMiddleware or unknown-type middlewares
        middlewares = middlewares.concat(
          userDefinedMiddlewares
            .filter((middleware) => !Helper.instanceOf(middleware, AfterHandlerMiddleware))
            .map((middleware) => (Helper.instanceOf(middleware, BeforeHandlerMiddleware) ? middleware.handle.bind(middleware) : middleware)),
        );
        middlewares = middlewares.concat([requestHandler, errorHandler]);
        // set AfterHandlerMiddleware middlewares
        middlewares = middlewares
          .concat(monitoringMiddlewares.after)
          .concat(
            userDefinedMiddlewares
              .filter((middleware) => Helper.instanceOf(middleware, AfterHandlerMiddleware))
              .map((middleware) => (Helper.instanceOf(middleware, AfterHandlerMiddleware) ? middleware.handle.bind(middleware) : middleware)),
          );

        router[method](apiPath, middlewares);
      });
    });

    app.use(this.getVariable('dino:server:path', '/'), router);
    let port = this.getVariable('dino:server:port', 3030);
    this.server = app.listen(port, () => {
      Logger.info(`application started, listening on port ${port}`);
    });
  }

  resolveDescriptorProvider() {
    let answer = null;
    try {
      Logger.info('try locate user defined descriptor provider from the application context');
      answer = this.applicationContext.resolve('descriptorProvider');
    } catch (e) {
      Logger.info('user defined descriptor provider not found using default file system based one');
      let DescriptorProvider = require(DESCRIPTOR_PROVIDER_MAP[this.getVariable('dino:descriptor:provider', 'filesystem')]);
      answer = new DescriptorProvider(this.environment);
    }
    return answer;
  }

  setupMonitoringEndpointsIfRequired(router) {
    let monitoringEnabled = this.getVariable('dino:server:monitoring:enabled', true);
    if (monitoringEnabled) {
      this.applicationStateMonitor = new ApplicationStateMonitor(this.applicationContext);
      router.use('/monitor', this.applicationStateMonitor.middleware(this.applicationContext));
    }
  }

  /**
   * Allows to define top level monitors. The returned obects contains two arrays
   * before and after middlewares, the before middlewares are added at the very start
   * of the middleware chain while the after are added at the very end of the middleware
   * chain, both in the order that are defined.
   *
   * @returns {Object}
   * ```{
   *    before: [],
   *    after: []
   *  }```
   */
  defineTopLevelMonitorsIfRequired() {
    const answer = { before: [], after: [] };
    let monitoringEnabled = this.getVariable('dino:server:monitoring:enabled', true);
    let co2MonitoringEnabled = this.getVariable('dino:server:monitoring:co2', true);
    if (monitoringEnabled) {
      if (co2MonitoringEnabled) {
        const co2Monitor = new CO2EmissionMonitor();
        this.addMonitor(co2Monitor);
        answer.before.push(co2Monitor.requestMiddleware.bind(co2Monitor));
        answer.after.push(co2Monitor.responseMiddleware.bind(co2Monitor));
      }
    }
    return answer;
  }

  /**
   * Allows to add a monitor
   * @param {Monitor} monitor
   */
  addMonitor(monitor) {
    if (this.applicationStateMonitor) {
      this.applicationStateMonitor.addMonitor(monitor);
    }
  }

  preDestroy() {
    if (this.server) {
      this.server.close();
    }
  }

  /**
   * Hook method allow to customize the express instance. by default it add the json middleware only.
   * This method can be overwritten in order to provide more tailored configuration of express instance.
   *
   * @param {any} express the express instance
   * @protected
   */
  extend(express) {
    // eslint-disable-line no-unused-vars
  }

  /**
   * Hook allows to provide an express error handler, by default a JSON error handler is used.
   * @protected
   */
  errorHandler() {
    return ErrorHandler.instance();
  }

  /**
   * Allows users to provide custom middlewares the middlewares will be inserted on the
   * chain as provided by the user but between the built-in API validation and route
   * handler and error handler.
   * @param {any} api the swagger API definition
   * @param {any} components the components OpenAPI definition
   * @returns {Array<Function>}
   *
   * @protected
   */
  middlewares(api, components) {
    return [].concat(this.getCache(api, components));
  }

  /**
   * Hook allow to define a custom request handler middleware. If no middleware is
   * returned the default will be used.
   * @param {any} api the OpenAPI definition
   * @param {function} responseValidator the validato for the response expected by the OpenAPI definition.
   * @returns {Function}
   *
   * @protected
   */
  requestHandler(api, responseValidator) {
    // eslint-disable-line no-unused-vars
    const routeHandler = new RouteHandler(this.applicationContext, api, responseValidator, this.environment);
    return routeHandler.handle.bind(routeHandler);
  }

  /**
   * Retrieve the value associated with the key or the default.
   * @param {String} key the key for the value to retrieve
   * @param {any} _default the default value
   *
   * @private
   */
  getVariable(key, _default) {
    return Helper.getVariable(this.environment, key, _default);
  }

  /**
   * Get the cache middleware configured as requested for the current route
   * @param {Aany} cache the cache configuration
   * @returns {Function} the cache middleware.
   *
   * @private
   */
  // eslint-disable-next-line no-unused-vars
  getCache(api, components) {
    Logger.debug(`configuring cache for API ${api.operationId}`);
    let answer = (req, res, next) => next();
    if (api.cache) {
      Logger.debug(`configuring cache ttl to ${api.cache.ttl}`);
      const cacheInstance = apicache.newInstance({
        debug: api.cache.debug || false,
        trackPerformance: api.cache.trackPerformance || true,
        isBypassable: api.cache.isBypassable || true,
      });
      answer = cacheInstance.middleware(api.cache.ttl || '30 minutes', this.cacheOnlySuccessResponsesForGet);
      this.addMonitor(new CachePerformanceMonitor().setApiName(api.operationId).setCacheInstance(cacheInstance));
    }
    return answer;
  }

  /**
   * Define a function that discriminate what responses should be cahced. In this case only responses
   * with 200 response code from GET requests.
   * @returns {Function} the cache toggle function
   *
   * @private
   */
  cacheOnlySuccessResponsesForGet(req, res) {
    Logger.debug(`${res.statusCode} ${req.method}`);
    return res.statusCode == 200 && req.method.toLowerCase() == 'get';
  }
}

/**
 * Configure express routing based on swagger configuration.
 *
 * @typedef RoutingConfigurer
 * @public
 */
module.exports = RoutingConfigurer;
